/*
 * SMPBasicBlock.h - <see below>.
 *
 * Copyright (c) 2000, 2001, 2010 - University of Virginia 
 *
 * This file is part of the Memory Error Detection System (MEDS) infrastructure.
 * This file may be used and modified for non-commercial purposes as long as 
 * all copyright, permission, and nonwarranty notices are preserved.  
 * Redistribution is prohibited without prior written consent from the University 
 * of Virginia.
 *
 * Please contact the authors for restrictions applying to commercial use.
 *
 * THIS SOURCE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: University of Virginia
 * e-mail: jwd@virginia.com
 * URL   : http://www.cs.virginia.edu/
 *
 * Additional copyrights 2010, 2011, 2014, 2015 by Zephyr Software LLC
 * e-mail: {clc,jwd}@zephyr-software.com
 * URL   : http://www.zephyr-software.com/
 *
 */

#ifndef SMPBASICBLOCK_H
#define SMPBASICBLOCK_H 1

// SMPBasicBlock.h
//
// This header defines the interfaces needed for analyzing basic blocks.

#include <list>
#include <set>
#include <string>

#include <cstddef>

class STARSCFGBlock; // forward declaration

#include "interfaces/STARSTypes.h"
#include "interfaces/SMPDBInterface.h"
#include "base/SMPDataFlowAnalysis.h"
#include "base/SMPInstr.h"
#include "base/SMPFunction.h"

// We can use less memory if we don't build explicit def-use chains.
#define STARS_BUILD_DEF_USE_CHAINS 0

// Aggressively propagate stack memory types until we see an indirect memory write
#define STARS_PROPAGATE_STACKMEMTYPE_UNTIL_INDIRECT 1

// Value for a block number before it is initialized.
#define SMP_BLOCKNUM_UNINIT (-1)

// Value for a pseudo-block that multiple RETURN blocks can be thought of as branching to.
#define SMP_BLOCKNUM_COMMON_RETURN (-2)

// Value for an SSA subscript number before it is initialized by SSA renaming.
#define SMP_SSA_UNINIT  (-1)

// Masks to set bits to use as booleans1 in SMPBasicBlock.
#define BLOCK_SET_PROCESSED 0x01
#define BLOCK_SET_CALLS_ALLOCA 0x02
#define BLOCK_SET_RETURNS 0x04
#define BLOCK_SET_SHARED_TAIL_CHUNK 0x08
#define BLOCK_SET_LIVE_IN_STALE 0x10
#define BLOCK_SET_UNREACHABLE_BLOCK 0x20
#define BLOCK_SET_FLAGS_DEAD_BEFORE_FIRST_KILL 0x40
#define BLOCK_SET_ALIASED_WRITE 0x80
#define BLOCK_RESET_PROCESSED 0xfe
#define BLOCK_RESET_CALLS_ALLOCA 0xfd
#define BLOCK_RESET_RETURNS 0xfb
#define BLOCK_RESET_SHARED_TAIL_CHUNK 0xf7
#define BLOCK_RESET_LIVE_IN_STALE 0xef
#define BLOCK_RESET_UNREACHABLE_BLOCK 0xdf
#define BLOCK_RESET_FLAGS_DEAD_BEFORE_FIRST_KILL 0xbf
#define BLOCK_RESET_ALIASED_WRITE 0x7f

// Masks to set bits to use as booleans2 in SMPBasicBlock.
#define BLOCK_SET_REACHES_OUT_STALE 0x01
#define BLOCK_SET_LOOP_TAIL_BLOCK 0x02
#define BLOCK_SET_LOOP_HEADER_BLOCK 0x04
#define BLOCK_SET_CALLS_UNSIGNED_ARG_FUNC 0x08
#define BLOCK_SET_SCCP_VISITED 0x10
#define BLOCK_SET_HAS_CALL_INST 0x20
#define BLOCK_SET_SINGLE_EXPRESSION 0x40
#define BLOCK_SET_OPTIMIZED_TOP_LOOP_TEST 0x80
#define BLOCK_RESET_REACHES_OUT_STALE 0xfe
#define BLOCK_RESET_LOOP_TAIL_BLOCK 0xfd
#define BLOCK_RESET_LOOP_HEADER_BLOCK 0xfb
#define BLOCK_RESET_CALLS_UNSIGNED_ARG_FUNC 0xf7
#define BLOCK_RESET_SCCP_VISITED 0xef
#define BLOCK_RESET_HAS_CALL_INST 0xdf
#define BLOCK_RESET_SINGLE_EXPRESSION 0xbf
#define BLOCK_RESET_OPTIMIZED_TOP_LOOP_TEST 0x7f

// Masks to set bits to use as booleans3 in SMPBasicBlock.
#define BLOCK_SET_SWITCH_DEFAULT_CASE 0x01
#define BLOCK_SET_HAS_INDIRECT_WRITE 0x02
#define BLOCK_SET_HAS_NORET_CALL 0x04
#define BLOCK_SET_DOUBLE_TAIL_BLOCK 0x08  // tail block for multiple loops, i.e. falls through to head of one, branches to head of other
#define BLOCK_SET_HAS_LOOPING_STRING_OPCODE 0x10
#define BLOCK_SET_UNUSED5 0x20
#define BLOCK_SET_UNUSED6 0x40
#define BLOCK_SET_UNUSED7 0x80
#define BLOCK_RESET_SWITCH_DEFAULT_CASE 0xfe
#define BLOCK_RESET_HAS_INDIRECT_WRITE 0xfd
#define BLOCK_RESET_HAS_NORET_CALL 0xfb
#define BLOCK_RESET_DOUBLE_TAIL_BLOCK 0xf7
#define BLOCK_RESET_HAS_LOOPING_STRING_OPCODE 0xef
#define BLOCK_RESET_UNUSED5 0xdf
#define BLOCK_RESET_UNUSED6 0xbf
#define BLOCK_RESET_UNUSED7 0x7f

// Class defining basic blocks.
class SMPBasicBlock {
public:
	// Constructors
	SMPBasicBlock(SMPFunction *Func, std::list<SMPInstr *>::iterator FirstInstr, std::list<SMPInstr *>::iterator LastInstr);

	// Operators
	int operator==(const SMPBasicBlock &rhs) const;

	// Get methods
	STARS_ea_t GetFirstAddr(void) const;
	STARS_ea_t GetLastAddr(void) const;
	inline int GetNumber(void) const { return BlockNum; };
	inline SMPBasicBlock *GetThisBlock(void) { return this; };
	inline SMPFunction *GetFunc(void) const { return MyFunc; };
#if 0
	inline STARS_sval_t GetIncomingStackPtrDelta(void) const { return IncomingStackPtrDelta; };
#endif
	inline std::vector<SMPInstr *>::iterator GetFirstInst(void) { return InstVec.begin(); };
	inline std::vector<SMPInstr *>::iterator GetLastInst(void) { return InstVec.end(); };
	inline std::vector<SMPInstr *>::const_iterator GetLastConstInst(void) const { return InstVec.cend(); };
	inline std::vector<SMPInstr *>::reverse_iterator GetRevInstBegin(void) { return InstVec.rbegin(); };
	inline std::vector<SMPInstr *>::reverse_iterator GetRevInstEnd(void) { return InstVec.rend(); }
	inline std::vector<SMPInstr *>::const_reverse_iterator GetRevInstCBegin(void) const { return InstVec.rbegin(); };
	inline std::vector<SMPInstr *>::const_reverse_iterator GetRevInstCEnd(void) const { return InstVec.rend(); };
	inline std::list<SMPBasicBlock *>::iterator GetFirstPred(void) {
		return Predecessors.begin();
	};
	inline std::list<SMPBasicBlock *>::iterator GetLastPred(void) {
		return Predecessors.end();
	};
	inline std::list<SMPBasicBlock *>::iterator GetFirstSucc(void) {
		return Successors.begin();
	};
	inline std::list<SMPBasicBlock *>::iterator GetLastSucc(void) {
		return Successors.end();
	};
	inline std::list<SMPBasicBlock *>::const_iterator GetFirstConstPred(void) const {
		return Predecessors.cbegin();
	};
	inline std::list<SMPBasicBlock *>::const_iterator GetLastConstPred(void) const {
		return Predecessors.cend();
	};
	inline std::list<SMPBasicBlock *>::const_iterator GetFirstConstSucc(void) const {
		return Successors.cbegin();
	};
	inline std::list<SMPBasicBlock *>::const_iterator GetLastConstSucc(void) const {
		return Successors.cend();
	};
	inline std::list<SMPBasicBlock *>::reverse_iterator GetRevSuccBegin(void) {
		return Successors.rbegin();
	};
	inline std::list<SMPBasicBlock *>::reverse_iterator GetRevSuccEnd(void) {
		return Successors.rend();
	};
	std::list<SMPBasicBlock *>::const_iterator GetFallThroughSucc(void) const; // Search successors for fall-through block, if any
	std::list<SMPBasicBlock *>::const_iterator GetCondNonFallThroughSucc(void) const; // Search successors for condition non-fall-through block, if any
	std::list<SMPBasicBlock *>::const_iterator GetCondOtherSucc(int NotThisBlockNum) const; // Search successors for conditional successor that is not #NotThisBlockNum
	inline std::size_t GetNumPreds(void) const { return Predecessors.size(); };
	inline std::size_t GetNumSuccessors(void) const { return Successors.size(); };
	inline int GetLoopHeaderNumber(void) const { return LoopHeaderNumber; };
	std::size_t GetBlockSize(void); // returns size in bytes
	std::set<STARSOpndTypePtr, LessOp>::iterator GetFirstLiveIn(void); // LiveInSet.begin(); recompute first if stale
	std::set<STARSOpndTypePtr, LessOp>::iterator GetLastLiveIn(void); // LiveInSet.end()
	std::set<STARSOpndTypePtr, LessOp>::iterator GetFirstLiveOut(void); // LiveOutSet.begin()
	std::set<STARSOpndTypePtr, LessOp>::iterator GetLastLiveOut(void); // LiveOutSet.end()
	std::set<STARSOpndTypePtr, LessOp>::iterator GetFirstVarKill(void); // KillSet.begin()
	std::set<STARSOpndTypePtr, LessOp>::iterator GetLastVarKill(void); // KillSet.end()
	std::set<STARSOpndTypePtr, LessOp>::iterator GetFirstUpExposed(void); // UpExposedSet.begin()
	std::set<STARSOpndTypePtr, LessOp>::iterator GetLastUpExposed(void); // UpExposedSet.end()
	inline std::size_t GetUpExposedSetSize(void) const { return UpExposedSet.size(); };
	inline STARSDefinitionSet::iterator GetFirstReachesIn(void) { return ReachesInSet.begin(); };
	inline STARSDefinitionSet::iterator GetLastReachesIn(void) { return ReachesInSet.end(); };
	inline STARSDefinitionSet::iterator GetFirstReachesOut(void) { return ReachesOutSet.begin(); };
	inline STARSDefinitionSet::iterator GetLastReachesOut(void) { return ReachesOutSet.end(); };
	inline STARSDefinitionSet::iterator GetFirstDownExposedDefn(void) { return DownExposedDefnSet.begin(); };
	inline STARSDefinitionSet::iterator GetLastDownExposedDefn(void) { return DownExposedDefnSet.end(); };
	std::set<STARSOpndTypePtr, LessOp>::iterator GetFirstLocalName(void); // LocalNames.begin()
	std::set<STARSOpndTypePtr, LessOp>::iterator GetLastLocalName(void); // LocalNames.end()
	std::set<int>::iterator GetFirstDomFrontier(void); // DomFrontier.begin()
	std::set<int>::iterator GetLastDomFrontier(void); // DomFrontier.end()
	std::set<SMPPhiFunction, LessPhi>::iterator GetFirstPhi(void); // PhiFunctions.begin()
	std::set<SMPPhiFunction, LessPhi>::iterator GetLastPhi(void); // PhiFunctions.end()
	int GetPredPosition(int BlockNum); // What is position # within Preds of BlockNum?
	std::set<SMPPhiFunction, LessPhi>::iterator FindPhi(const STARSOpndTypePtr &FindOp);
	std::set<STARSOpndTypePtr, LessOp>::iterator FindLocalName(const STARSOpndTypePtr &FindOp);
	inline std::map<int, struct STARS_SCCP_Const_Struct>::iterator FindLocalConstValue(int DefHashValue) { return ConstantLocalDefs.find(DefHashValue); };
	inline std::map<int, struct STARS_SCCP_Const_Struct>::iterator GetLastLocalConstValueIter(void) { return ConstantLocalDefs.end(); };
#if 0
	// Given a USE operand and the address of its instr, return DEF from DU chains or Phi func.
	std::set<DefOrUse, LessDefUse>::iterator GetDefFromOpAddr(STARSOpndTypePtr UseOp, STARS_ea_t InstAddr, int SSANum);
#endif
	std::set<DefOrUse, LessDefUse>::iterator GetGlobalDefIterFromDefAddr(const STARSOpndTypePtr &DefOp, STARS_ea_t DefAddr);
	STARS_ea_t GetDefAddrFromUseAddr(const STARSOpndTypePtr &UseOp, STARS_ea_t InstAddr, int SSANum, bool LocalName); // return DefAddr, block # for Phi Defs, or UpExposed address.
	STARS_ea_t GetUltimateDefAddr(const STARSOpndTypePtr &UseOp, STARS_ea_t InstAddr, int SSANum, bool LocalName, bool RecurseOnMoves, STARSOpndTypePtr &DefMoveOp); // trace into predecessors if necessary
	SMPInstr *FindInstr(STARS_ea_t InstAddr);
	std::vector<SMPInstr *>::iterator GetInstIterFromAddr(STARS_ea_t InstAddr);
	std::vector<SMPInstr *>::const_iterator GetInstConstIterFromAddr(STARS_ea_t InstAddr) const;
	inline SMPInstr *GetInstFromIndex(size_t VecIndex) const { return InstVec.at(VecIndex); };
	size_t GetIndexFromInstAddr(STARS_ea_t InstAddr) const; // Get the InstVec index for the instruction at InstAddr. Assert if not found.
	SMPInstr *FindBranchCompareOrTest(SMPoperator &BranchOperator, STARS_ea_t &DecAddr) const; // Find the compare or test that sets the flags used in COND_BRANCH at end of block, else NULL; If decrement found instead, fill DecAddr
	inline STARS_sval_t GetOutgoingStackDelta(void) const { return OutgoingStackDelta; };
	ControlFlowType GetLastInstCFType(void) const;  // Get the enum ControlFlowType for the last inst

	// Six methods to get values from the maps of local reg/SSA to FG info.
	//  For global names, see corresponding methods in SMPFunction.
	unsigned short GetDefSignMiscInfo(int DefHashValue);
	unsigned short GetUseSignMiscInfo(int UseHashValue);
	unsigned short GetDefWidthTypeInfo(int DefHashValue);
	unsigned short GetUseWidthTypeInfo(int UseHashValue);
	struct FineGrainedInfo GetDefFGInfo(int DefHashValue);
	struct FineGrainedInfo GetUseFGInfo(int UseHashValue);

	// Set methods
#if 0
	inline void SetIncomingStackPtrDelta(STARS_sval_t delta) { IncomingStackPtrDelta = delta; };
#endif
	inline void SetProcessed(bool flag) { flag ? (booleans1 |= BLOCK_SET_PROCESSED) : (booleans1 &= BLOCK_RESET_PROCESSED); };
	inline void SetMaybeAliased(bool flag) { flag ? (booleans1 |= BLOCK_SET_ALIASED_WRITE) : (booleans1 &= BLOCK_RESET_ALIASED_WRITE); };
	inline void SetReturns(bool flag) { flag ? (booleans1 |= BLOCK_SET_RETURNS) : (booleans1 &= BLOCK_RESET_RETURNS); };
	inline void SetCallsAlloca(void) { booleans1 |= BLOCK_SET_CALLS_ALLOCA; }
	inline void SetShared(void) { booleans1 |= BLOCK_SET_SHARED_TAIL_CHUNK; };
	inline void ResetShared(void) { booleans1 &= BLOCK_RESET_SHARED_TAIL_CHUNK; };
	inline void SetLiveInStale(bool flag) { flag ? (booleans1 |= BLOCK_SET_LIVE_IN_STALE) : (booleans1 &= BLOCK_RESET_LIVE_IN_STALE); };
	inline void SetUnreachableBlock(bool flag) { flag ? (booleans1 |= BLOCK_SET_UNREACHABLE_BLOCK) : (booleans1 &= BLOCK_RESET_UNREACHABLE_BLOCK); };
	inline void SetFlagsDeadBeforeFirstKill(bool flag) { flag ? (booleans1 |= BLOCK_SET_FLAGS_DEAD_BEFORE_FIRST_KILL) : (booleans1 &= BLOCK_RESET_FLAGS_DEAD_BEFORE_FIRST_KILL); };
	inline void SetReachesOutStale(void) { booleans2 |= BLOCK_SET_REACHES_OUT_STALE; };
	inline void SetLoopHeaderBlock(void) { booleans2 |= BLOCK_SET_LOOP_HEADER_BLOCK; };
	inline void ResetLoopHeaderBlock(void) { booleans2 &= BLOCK_RESET_LOOP_HEADER_BLOCK; };
	inline void SetLoopHeaderBlockNumberForTailBlock(const int HeadBlockNum) { LoopHeaderNumber = HeadBlockNum; };
	inline void SetLoopTailBlock(void) { booleans2 |= BLOCK_SET_LOOP_TAIL_BLOCK; };
	inline void ResetLoopTailBlock(void) { booleans2 &= BLOCK_RESET_LOOP_TAIL_BLOCK; };
	inline void SetCallsUnsignedArgFunc(void) { booleans2 |= BLOCK_SET_CALLS_UNSIGNED_ARG_FUNC; };
	inline void SetSCCPVisited(bool flag) { flag ? (booleans2 |= BLOCK_SET_SCCP_VISITED) : (booleans2 &= BLOCK_RESET_SCCP_VISITED); };
	inline void SetHasCallInst(void) { booleans2 |= BLOCK_SET_HAS_CALL_INST; };
	inline void SetSingleExpression(void) { booleans2 |= BLOCK_SET_SINGLE_EXPRESSION; };
	inline void SetOptimizedTopLoopTest(void) { booleans2 |= BLOCK_SET_OPTIMIZED_TOP_LOOP_TEST; };
	inline void SetSwitchDefaultCase(void) { booleans3 |= BLOCK_SET_SWITCH_DEFAULT_CASE; };
	inline void SetHasIndirectMemWrite(void) { booleans3 |= BLOCK_SET_HAS_INDIRECT_WRITE; };
	inline void SetHasNonReturningCall(void) { booleans3 |= BLOCK_SET_HAS_NORET_CALL; };
	inline void SetLoopDoubleTailBlock(void) { booleans3 |= BLOCK_SET_DOUBLE_TAIL_BLOCK; };
	inline void SetHasLoopingStringOpcode(void) { booleans3 |= BLOCK_SET_HAS_LOOPING_STRING_OPCODE; };
	inline void SetNumber(int Num) { BlockNum = Num; };
	inline void SetOutgoingStackDelta(STARS_sval_t OutDelta) { OutgoingStackDelta = OutDelta; };
	void LinkToPred(SMPBasicBlock *Predecessor);
	void LinkToSucc(SMPBasicBlock *Successor);
	inline void AddLiveIn(STARSOpndTypePtr LiveOp) { LiveInSet.insert(LiveOp); };
	inline void AddLiveOut(STARSOpndTypePtr LiveOp) { LiveOutSet.insert(LiveOp); };
	inline void AddVarKill(STARSOpndTypePtr KillOp) { KillSet.insert(KillOp); };
	inline void AddUpExposed(STARSOpndTypePtr UseOp) { UpExposedSet.insert(UseOp); };
	void AddReachesIn(std::pair<STARSOpndTypePtr, STARS_ea_t> ReachInDefn);
	void AddReachesOut(std::pair<STARSOpndTypePtr, STARS_ea_t> ReachOutDefn);
	inline void AddDownExposedDefn(std::pair<STARSOpndTypePtr, STARS_ea_t> DEDefn) { DownExposedDefnSet.insert(DEDefn); };
	inline void ClearPhiFunctions(void) { PhiFunctions.clear(); };
	bool ErasePhi(const STARSOpndTypePtr &OldOp); // erase() from phi function list
	void ErasePred(STARS_ea_t FirstBlockAddr); // erase() block starting at FirstBlockAddr from Preds list
	void EraseSucc(STARS_ea_t FirstBlockAddr); // erase() block starting at FirstBlockAddr from Successors list
	bool AddPhi(SMPPhiFunction NewPhi);
	bool SetPhiDefSSA(const STARSOpndTypePtr &DefOp, int SSANum);
	bool SetPhiUseSSA(const STARSOpndTypePtr &UseOp, std::size_t index, int SSANum);
	std::set<SMPPhiFunction, LessPhi>::iterator SetPhiDefType(const STARSOpndTypePtr &DefOp, SMPOperandType Type);
	bool SetPhiUseType(const STARSOpndTypePtr &DefOp, std::size_t index, SMPOperandType Type);
	std::set<SMPPhiFunction, LessPhi>::iterator SetPhiDefMetadata(const STARSOpndTypePtr &DefOp, SMPMetadataType Status);
	std::map<int, struct STARS_SCCP_Const_Struct>::iterator InsertLocalConstValue(int DefHashValue, struct STARS_SCCP_Const_Struct NewConstEntry); 
		// Insert SCCP value for local name; change if already found.

	// Six methods to set values into the maps of local reg/SSA to FG info.
	//  For global names, see corresponding methods in SMPFunction.
	void UpdateDefSignMiscInfo(int DefHashValue, unsigned short NewInfo);
	void UpdateUseSignMiscInfo(int UseHashValue, unsigned short NewInfo);
	void UpdateDefWidthTypeInfo(int DefHashValue, unsigned short NewInfo);
	void UpdateUseWidthTypeInfo(int UseHashValue, unsigned short NewInfo);
	void UpdateDefFGInfo(int DefHashValue, struct FineGrainedInfo NewFG);
	void UpdateUseFGInfo(int UseHashValue, struct FineGrainedInfo NewFG);

	void ClearDefSignedness(int DefHashValue);
	void InitPointerCopies(std::size_t LocalSSALimit);  // initialize InArgPointerCopies container
	void SetInArgPointerCopy(const STARSOpndTypePtr &CurrOp, int SSANum); // set bit in InArgPointerCopies

	// Query methods
	inline bool IsProcessed(void) const { return (booleans1 & BLOCK_SET_PROCESSED); };
	inline bool HasIndirectJump(void) { return (INDIR_JUMP == (*(GetRevInstBegin()))->GetDataFlowType()); };
	inline bool HasDirectJump(void) { return (JUMP == (*(GetRevInstBegin()))->GetDataFlowType()); };
	inline bool EndsWithTailCall(void) { return (*(GetRevInstBegin()))->IsTailCall(); };
	inline bool CallsAlloca(void) const { return (booleans1 & BLOCK_SET_CALLS_ALLOCA); };
	inline bool HasReturn(void) const { return (booleans1 & BLOCK_SET_RETURNS); };
	inline bool IsShared(void) const { return (booleans1 & BLOCK_SET_SHARED_TAIL_CHUNK); };
	inline bool IsLiveInStale(void) const { return (booleans1 & BLOCK_SET_LIVE_IN_STALE); };
	inline bool IsUnreachableBlock(void) const { return (booleans1 & BLOCK_SET_UNREACHABLE_BLOCK); };
#if 0
	inline bool AreFlagsDeadAfterLastUse(void) const { return (booleans1 & BLOCK_SET_FLAGS_DEAD_AFTER_LAST_USE); };
#endif
	inline bool AreFlagsDeadBeforeFirstKill(void) const { return (booleans1 & BLOCK_SET_FLAGS_DEAD_BEFORE_FIRST_KILL); };
	bool AllNops(void); // Are all instructions in the block no-ops?
	inline bool MaybeAliasedWrite(void) const { return (booleans1 & BLOCK_SET_ALIASED_WRITE); };
	inline bool IsReachesOutStale(void) const { return (booleans2 & BLOCK_SET_REACHES_OUT_STALE); };
	inline bool IsLoopTailBlock(void) const { return (booleans2 & BLOCK_SET_LOOP_TAIL_BLOCK); };
	inline bool IsLoopHeaderBlock(void) const { return (booleans2 & BLOCK_SET_LOOP_HEADER_BLOCK); };
	inline bool IsSingleExpression(void) const { return (booleans2 & BLOCK_SET_SINGLE_EXPRESSION); };
	inline bool CallsUnsignedArgFunc(void) const { return (booleans2 & BLOCK_SET_CALLS_UNSIGNED_ARG_FUNC); };
	inline bool IsSCCPVisited(void) const { return (booleans2 & BLOCK_SET_SCCP_VISITED); };
	inline bool HasCallInstruction(void) const { return (booleans2 & BLOCK_SET_HAS_CALL_INST); };
	inline bool IsOptimizedTopLoopTest(void) const { return (booleans2 & BLOCK_SET_OPTIMIZED_TOP_LOOP_TEST); };
	inline bool IsSwitchDefaultCase(void) const { return (booleans3 & BLOCK_SET_SWITCH_DEFAULT_CASE); };
	inline bool HasIndirectMemWrite(void) const { return (booleans3 & BLOCK_SET_HAS_INDIRECT_WRITE); };
	inline bool HasNonReturningCall(void) const { return (booleans3 & BLOCK_SET_HAS_NORET_CALL); };
	inline bool IsDoubleLoopTailBlock(void) const { return (booleans3 & BLOCK_SET_DOUBLE_TAIL_BLOCK); };
	inline bool HasLoopingStringOpcode(void) const { return (booleans3 & BLOCK_SET_HAS_LOOPING_STRING_OPCODE); };
	inline bool IsSelfLoop(void) const { return (IsLoopHeaderBlock() && IsLoopTailBlock()); };
	bool IsBlockPred(int BlockNum) const; // Is BlockNum a predecessor block of this block
	bool HasLoopHeadAsSuccessor(int &HeadBlockNum) const; // block has HeadBlockNum as successor, which is a loop header
	bool HasLoopHeadWithInvertedExitAsSuccessor(void) const; // block has loop head successor that ends with INVERTED_LOOP_EXIT
	bool IsLiveIn(const STARSOpndTypePtr &CurrOp);
	inline bool IsLiveOut(const STARSOpndTypePtr &CurrOp) const {
		return (LiveOutSet.end() != LiveOutSet.find(CurrOp));
	}
	inline bool IsUpExposed(const STARSOpndTypePtr &CurrOp) const {
		return (UpExposedSet.end() != UpExposedSet.find(CurrOp));
	}
	inline bool IsVarKill(const STARSOpndTypePtr &CurrOp) const {
		return (KillSet.end() != KillSet.find(CurrOp));
	}
	bool MDAlreadyKilled(const STARSOpndTypePtr &Opnd1) const; // Was STARSOpndTypePtr killed by something already in KillSet?
	inline bool IsReachesIn(std::pair<STARSOpndTypePtr, STARS_ea_t> CurrDefn) const {
		return (ReachesInSet.end() != ReachesInSet.find(CurrDefn));
	}
	inline bool IsReachesOut(std::pair<STARSOpndTypePtr, STARS_ea_t> CurrDefn) const {
		return (ReachesOutSet.end() != ReachesOutSet.find(CurrDefn));
	}
	inline bool IsDownExposedDefn(std::pair<STARSOpndTypePtr, STARS_ea_t> CurrDefn) const {
		return (DownExposedDefnSet.end() != DownExposedDefnSet.find(CurrDefn));
	}
	bool IsLocalName(const STARSOpndTypePtr &CurrOp) const;
	bool IsAddrInBlock(const STARS_ea_t InstAddr) const;
	bool HasConditionalBranch(void) const; // true if (last) instruction is conditional branch
	bool HasMemoryWrite(void) const; // Writes to stack or has indirect write
	bool IsInArgPointerCopy(const STARSOpndTypePtr &CurrOp, int SSANum) const; // found in local InArgPointerCopies?
	bool IsOpDestTruncatedWrite(const STARSOpndTypePtr &DefOp, int DefSSANum, STARS_ea_t DefAddr); // Does DefOp get written out in truncated form (lower bits only)?

	// Printing methods
	void Dump(void);
	void EmitSPARKAdaForFallThroughInsts(FILE *BodyFile); // Call SMPInstr::EmitSPARKAda() until we hit a control flow inst.
	void EmitSPARKAdaForAllInsts(FILE *BodyFile); // Call SMPInstr::EmitSPARKAda() for all insts in block.
	void EmitSPARKAdaForExpression(FILE *BodyFile); // For SingleExpression block, emit SPARK Ada expression
	bool EmitSPARKAdaForExprUse(FILE *BodyFile, std::vector<SMPInstr *>::const_reverse_iterator RevInstIter, STARSOpndTypePtr UseOp, int UseSSANum); // Emit SPARK expression for UseOp at RevInstIter

	// Analysis methods
	bool AllPredecessorsNumbered(void);
	bool AllPredecessorsProcessed(void);
	void DepthFirstMark(bool MarkIBTBlocks); // Depth-first traversal, mark as processed.
	STARS_ea_t GetUltimateOperandSource(STARS_ea_t UseAddr, const STARSOpndTypePtr &UseOp, int UseSSANum); // trace through move and Phi defs to some defining operation.
	void Analyze();
	void InitKilledExposed(bool UseFP); // Initialize KilledSet and UpExposedSet
	bool UpdateLiveOut(void); // Iterate once on updating LiveOutSet; return true if changed
	void SetLiveOutSSANums(void); // Set the SSA numbers for the DEFs in the LiveOut set.
	void AddToDomFrontier(int); // Add RPO block number to DomFrontier set.
	void SetLocalNames(void); // Fill the LocalNames member set
	void SSALocalRenumber(void); // Renumber references to local names
	bool DoesDefControlLoopFlow(const STARSOpndTypePtr &DefOp, int DefSSANum, STARS_ea_t DefAddr); // Does DefOp/DefSSANum control loop-back or loop-exit branches?
	void FreeReachingDefsMemory(void); // Free memory for reaching defs sets after they are no longer needed
	void FreeSSAMemory(void); // After SSA #s are in DEFs and USEs, free SSA data structures.
	void FreeUnusedMemory4(void); // After loop 4 (type inference) in SMPProgram::Analyze(), free memory
	bool IsMemWriteUnsafe(STARS_ea_t InstID, bool FirstBlock, STARSDefUseSet &DEFsToSearch); // Are any DEFs in search set of unsafe-write origin? 
	bool IsGlobalNameUsedLiveIn(const STARSOpndTypePtr &GlobalOp); // Is GlobalOp used as a LiveIn value anywhere in this block, including as a Phi USE?
	bool IsGlobalRegDead(STARS_ea_t InstAddr, const STARSOpndTypePtr &Operand, unsigned int RegIndex); // Is global reg dead at InstAddr?
	bool IsRegDead(STARS_ea_t InstAddr, STARS_regnum_t RegNo); // Is local reg dead at InstAddr?
	void MarkDeadRegs(void); // Find dead registers for each mmStrata-instrumented instruction
	bool IsDefDead(STARS_ea_t DefAddr, const STARSOpndTypePtr &DefOp); // Does DefOp at DefAddr never get used?
	bool AreFlagsLiveAfterInst(STARS_ea_t InstAddr, std::set<int> &LiveFlagRegs); // Return true and produce set if any flags are live out of inst at InstAddr.
	bool IsDefInvolvedInAddition(STARS_ea_t DefAddr, const STARSOpndTypePtr &DefOp, STARS_ea_t &AdditionAddr); // true if DefOp at DefAddr is used in or redefined by addition; pass back AdditionAddr
	bool IsUseCopiedAndShiftedRight(const STARSOpndTypePtr &UseOp, STARS_ea_t UseAddr); // Does a copy of UseOp get shifted right in this block?
	bool DoesDefReachBlockEnd(STARS_ea_t DefAddr, const STARSOpndTypePtr &DefOp, int DefSSANum, std::set<int> &NonEscapingRegisterHashes); // Does value of DefOp/DefSSANum reach block end in any DEF at all?
	bool IsDefOnlyUsedAsAddressReg(STARS_ea_t DefAddr, const STARSOpndTypePtr &DefOp, std::size_t LoopNum, std::size_t &AddrUseCounter, SMPOperandType &MemType); // DefOp/DefSSANum is only used in LoopNum (outside of DefAddr) as address reg or Phi USE
	bool IsDefUsedInLoopCompareAndBranch(STARS_ea_t DefAddr, const STARSOpndTypePtr &DefOp, int DefSSANum, bool &Signed); // Is DefOp/DefAddr used in a loop back edge compare-and-branch?
	bool FindSingleExpression(void); // return true so caller can SetSingleExpression() if block is only an expr for a compare and branch
	bool PropagateLocalDefType(const STARSOpndTypePtr &DefOp, SMPOperandType DefType, STARS_ea_t DefAddr, int SSANum, bool IsMemOp, bool PointerOverride = false); // to all uses
	bool InferLocalDefType(const STARSOpndTypePtr &DefOp, unsigned int LocIndex, STARS_ea_t DefAddr); // Can DEF type be inferred from all USEs?
	void InferGlobalDefType(const STARSOpndTypePtr &DefOp, int SSANum, STARS_ea_t DefAddr, bool &FoundNumeric, bool &FouondCodePtr, bool &FoundPointer,
		bool &FoundUnknown, bool &FoundUninit, SMPOperandType &PtrType);  // Can DEF type be inferred from all USEs? Recursive helper for SMPFunction::InferGlobal...
	bool PropagateGlobalDefType(const STARSOpndTypePtr &DefOp, SMPOperandType DefType, int SSANum, bool IsMemOp, bool PointerOverride, bool FirstBlock = true); // to all uses
	bool InferAllPhiDefTypes(void); // infer, propagate to all uses
	bool PropagateLocalMetadata(const STARSOpndTypePtr &UseOp, SMPMetadataType Status, int SSANum, STARS_ea_t UseAddr);
	bool FindRedundantLocalMetadata(bool SafeFunc); // consecutive DEFs, same type => redundant
	bool HasUseAfterIndWrite(const STARSOpndTypePtr &MemDefOp, std::vector<SMPInstr *>::iterator InstIter, bool &IndWriteSeen, bool &ReDef); // MemDefOp starting at InstIter has DU chain overlapping an indirect mem write
	void MarkBranchSignedness(void); // If branch at block end is signed/unsigned, propagate to operands that set flags before it.
	void PropagatePhiFGInfo(void); // Ensure that the Phi DEFs have all the fine grained info from all Phi USEs.
	void PropagateBranchSignedness(STARS_ea_t DefAddr, const STARSOpndTypePtr &SearchOp, unsigned short SignMask);
	bool PropagateDEFSignedness(void); // propagate DEF signedness to USE if USE has no signedness info.
	void MarkUnsignedArgs(STARS_ea_t CallAddr, unsigned int ArgPosBits); // backtrack from CallAddr and mark unsigned args based on bitset
	void MarkPointerArgs(STARS_ea_t CallAddr, unsigned int ArgPosBits); // backtrack from CallAddr and mark POINTER args based on bitset
	bool ComputeReachesInSet(void); // return true if changes were made to the ReachesInSet
	bool ComputeReachesOutSet(void); // return true if changes were made to the ReachesOutSet
	void UpdateDownExposedDefs(STARSOpndTypePtr DefOp, STARS_ea_t InstAddr); // Add DefOp, remove previous defs of DefOp that now do not reach the end of the block.

	void SCCPEvaluateConstants(enum STARSBranchConst &BranchEval, std::list<std::pair<int, int> > &CFGWorkList, std::list<std::pair<int, int> > &SSAWorkList); // Evaluate constants for all instructions as part of SCCP algorithm at SMPFunction level.
	void SCCPGlobalPropagationHelper(const STARSOpndTypePtr &GlobalOp, int DefSSANum, std::vector<STARSBitSet> ExecutedEdgeBitSet, std::list<std::pair<int, int> > &CFGWorkList, std::list<std::pair<int, int> > &SSAWorkList); // Propagate GlobalOp/DefSSANum const value
	void SCCPHandleSuccessors(enum STARSBranchConst &BranchEval, std::list<std::pair<int, int> > &CFGWorkList); // Based on BranchEval, push successors onto CFGWorkList
	void SCCPNullifyUnreachableBlock(void); // Patch binary to make this block into nops.
	void SCCPGatherStatistics(void); // Gather statistics on SCCP effectiveness

	// Find the stack target of a call to memset() and the size in bytes of the memset() region.
	//  If the memset() target is not on the stack, or the size is not a constant, return false.
	bool AnalyzeMemSet(STARS_ea_t MemSetAddr, STARSOpndTypePtr &MemSetTarget, std::size_t &MemSetSize, int &SignedOffset, bool &FPRelativeTarget);
	
	bool IsInfiniteSelfLoop(void); // return true if block loops infinitely to itself
	bool IsBenignOverflowDEF(const STARSOpndTypePtr &DefOp, int DefSSANum, STARS_ea_t DefAddr, bool UnderflowOpcode, int &IdiomCode); // Do we not care if DEF overflowed, due to how it is used?
	bool IsDEFWrittenWithTruncation(const STARSOpndTypePtr &DefOp, int DefSSANum, STARS_ea_t DefAddr, STARS_ea_t TruncAddr, bool PhiDEF); // Is subword of DefOp written later, before TruncAddr?
	bool IsBenignTruncationDEF(const STARSOpndTypePtr &DefOp, int DefSSANum, STARS_ea_t DefAddr, int &IdiomCode); // Do we not care if DEF overflowed, due to how it is used?
	bool IsDefMaskedOrShiftedRight(const STARSOpndTypePtr &DefOp, int DefSSANum, STARS_ea_t DefAddr); // Does DefOp+DefSSANum get subreg masked or right shifted before it gets re-DEFined?
	bool IsCriticalSink(const STARSOpndTypePtr &DefOp, int DefSSANum, std::string &SinkString, bool &FoundAnyCall); // What is ultimate use of DEF that might have integer error?
	bool IsStackOpNextUsedWithSignedness(const STARSOpndTypePtr &StackDefOp, STARS_ea_t DefAddr, int SSANum); // Is next use of stack DEF a sign- or zero-extended load?
	void SuppressAnnotOnSignChangingAddition(const STARSOpndTypePtr &DefOp, int DefSSANum, STARS_ea_t DefAddr); // Find inc/add that makes small negative back into positive
	void AnalyzePrepForNumericAnnotations(void); // Last-minute prep before emitting numeric annotations
	// Is DefOp+DefSSANum (from WorkList head) at DefAddr used as address reg or as source operand in unsafe memory write?
	bool IsDefUsedInUnsafeMemWrite(std::list<std::pair<std::pair<STARSOpndTypePtr, int>, STARS_ea_t> > &WorkList, const STARSOpndTypePtr &DefOp, int DefSSANum, STARS_ea_t DefAddr);

	STARS_sval_t ComputeStackAdjustmentAfterCall(STARS_ea_t CallAddr); // find stack pointer adjustment code after call, return stack delta or zero
	STARS_ea_t FindFixedCallFallThrough(void) const; // Find the push operand that passed the fall-through address to a fixed call
	STARS_ea_t FindCallInstFallThrough(STARS_ea_t CallAddr) const; // Find inst after CallAddr; grab from successor if needed

private:
	// Data
	STARS_ea_t FirstAddr;
	int BlockNum;  // Number for block ordering algorithms
	int LoopHeaderNumber; // Number of block that is loop header for this block, if this is a tail block
	SMPFunction *MyFunc; // function containing this block
	STARS_sval_t OutgoingStackDelta; // Incoming stack ptr delta for all successor blocks
#if 0
	STARS_sval_t IncomingStackPtrDelta; // Incoming stack pointer delta for this block
#endif
	unsigned char booleans1; // hold bools as bits
#if 0  // now in booleans1
	bool Processed; // Set to true to prevent infinite recursions in CFG traversals

	// cached query results
	bool IndirectJump; // contains an indirect jump instruction
	bool Returns;      // contains a return instruction
	bool SharedTailChunk; // is part of a code chunk shared among functions
	bool IsLiveInStale; // Has LiveOutSet changed since LiveInSet was computed?
	bool FlagsDeadAfterLastUse; // flags global, dead after last use in the block
	bool FlagsDeadBeforeFirstKill; // flags global, dead before first DEF in block
	bool AliasedWrite; // A memory write in the block may be aliased by other writes
#endif
	unsigned char booleans2; // hold bools as bits
#if 0 // now in booleans2
	bool ReachesOutStale; // ReachesIn has been updated since last ReachesOut update
	bool LoopTailBlock; // block has back edge to loop header block
#endif
	unsigned char booleans3;
#if 0
	bool SwitchDefaultCase; // first block in default case of a switch statement
#endif

	std::vector<SMPInstr *> InstVec; //NOTE: Copies of pointers from SMPFunction::Instrs.
	std::map<STARS_ea_t, size_t> AddrInstMap; // map InstAddr to InstVec index.
	std::list<SMPBasicBlock *> Predecessors;
	std::list<SMPBasicBlock *> Successors;
		// Four sets used in live variable analysis
	std::set<STARSOpndTypePtr, LessOp> KillSet;   // variables killed in this block
	std::set<STARSOpndTypePtr, LessOp> UpExposedSet; // upward exposed variables in this block
	std::set<STARSOpndTypePtr, LessOp> LiveOutSet; // Live-Out variables in this block
	std::set<STARSOpndTypePtr, LessOp> LiveInSet; // contribution to predecessor's live-out iteration
		// Three sets used in reaching definitions analysis
	STARSDefinitionSet ReachesInSet; // definitions live on entry to block
	STARSDefinitionSet ReachesOutSet; // definitions live on exit from block, either passing through or DEFed inside block
	STARSDefinitionSet DownExposedDefnSet; // definitions within block that are live on exit
		// SSA data structures
	std::set<int> DomFrontier; // Dominance frontier for block, as set of RPO block numbers
	std::set<SMPPhiFunction, LessPhi> PhiFunctions; // SSA incoming edge phi functions
	std::set<STARSOpndTypePtr, LessOp> LocalNames; // non-global names referenced in this block
	std::vector<STARSBitSet> InArgPointerCopies; // indexed by LocalNameIndex, bitset indexed by SSA #, bit set to 1 => copy of InArg of CODEPTR or POINTER type
#if STARS_BUILD_DEF_USE_CHAINS
	SMPCompleteDUChains LocalDUChains; // def-use chains for local names
	SMPCompleteDUChains GlobalDUChains; // def-use chains for global names
		// NOTE: The GlobalChains.ChainsByName.at(GlobalIndex).DUChains are indexed
		//  starting at zero. The indices into the DUChains have nothing to do
		//  with the SSA Numbers, unlike the LocalDUChains.
#endif

	std::map<int, struct FineGrainedInfo> LocalDefFGInfoBySSA; // map hash of local name & SSANum to DEF FG info.
		// NOTE: We are currently limiting this map to registers, not all local names.

	std::map<int, struct FineGrainedInfo> LocalUseFGInfoBySSA; // map hash of local name & SSANum to USE FG info.
		// NOTE: We are currently limiting this map to registers, not all local names.

	std::map<int, struct STARS_SCCP_Const_Struct> ConstantLocalDefs; // map hash of local name & SSANum to constant value from SCCP.
		// NOTE: We are currently limiting this map to registers, not all local names. 

	// Methods
	inline void ResetReachesOutStale(void) { booleans2 &= BLOCK_RESET_REACHES_OUT_STALE; };
	void ComputeLiveInSet(void); // compute or recompute LiveIn set
	std::set<SMPPhiFunction, LessPhi>::iterator InferPhiDefType(std::set<SMPPhiFunction, LessPhi>::iterator DefPhi, bool &changed); // infer, propagate to all uses
#if STARS_BUILD_DEF_USE_CHAINS
	unsigned int GetLocalDUIndex(STARSOpndTypePtr DefOp, int SSANum);
	unsigned int GetGlobalDUIndex(STARSOpndTypePtr DefOp, STARS_ea_t DefAddr);
#endif
	bool UseHasCriticalSink(STARSOpndTypePtr DefOp, int DefSSANum, STARSOpndTypePtr DefOp2, int DefSSANum2, std::string &SinkString, bool &FoundAnyCall); // DefOp or DefOp2 is passed as arg to critical system or lib call
	bool IsBranchSignednessUnreliable(void); // Is branch signedness misleading due to odd code generation idiom or hand coding?
}; 

// Class to hold a tree expression of basic blocks that are just conditions, e.g. for compound
//  short-circuit conditional statements. Expressions are built up as conditional blocks are coalesced.

class STARSCondExpr {
public:
	// constructor
	STARSCondExpr() : ShortCircuitOper(SMP_NULL_OPERATOR), LeftExpr(NULL), RightExpr(NULL), NegateLeftExpr(false),
		LeftBlockNum(SMP_BLOCKNUM_UNINIT), RightBlockNum(SMP_BLOCKNUM_UNINIT), NonFallThroughBlockNum(SMP_BLOCKNUM_UNINIT),
		FallThroughBlockNum(SMP_BLOCKNUM_UNINIT) {};

	// Get methods
	SMPoperator GetOperator(void) const { return ShortCircuitOper; };
	STARSCondExpr *GetLeftExpr(void) const { return LeftExpr; };
	STARSCondExpr *GetRightExpr(void) const { return RightExpr; };
	int GetLeftBlockNum(void) const { return LeftBlockNum; };
	int GetRightBlockNum(void) const { return RightBlockNum; };
	int GetFallThroughBlockNum(void) const { return FallThroughBlockNum; };
	int GetNonFallThroughBlockNum(void) const { return NonFallThroughBlockNum; };

	// Set methods
	void SetOperator(SMPoperator Oper) { ShortCircuitOper = Oper; };
	void SetLeftExpr(STARSCondExpr *Expr, bool negated) { LeftExpr = Expr; NegateLeftExpr = negated; };
	void SetRightExpr(STARSCondExpr *Expr) { RightExpr = Expr; };
	void SetLeftExprBlockNum(int BlockNum, bool negated) { LeftBlockNum = BlockNum; NegateLeftExpr = negated; };
	void SetRightExprBlockNum(int BlockNum) { RightBlockNum = BlockNum; };
	void SetFTBlockNum(int BlockNum) { FallThroughBlockNum = BlockNum; };
	void SetNFTBlockNum(int BlockNum) { NonFallThroughBlockNum = BlockNum; };
	void SetCFGBlock(STARSCFGBlock *MyBlock) { MyCFGBlock = MyBlock; };

	// Printing methods
	void EmitSPARKAdaShortCircuitExpr(FILE *BodyFile); // Emit SPARK Ada short circuit expr for entire expression tree.

private:
	SMPoperator ShortCircuitOper;  // always SMP_LOGICAL_OR or SMP_LOGICAL_AND for our purposes, except initialized to SMP_NULL_OPERATOR
	STARSCondExpr *LeftExpr;  // NULL if left expr is just a block number
	STARSCondExpr *RightExpr; // NULL if right expr is just a block number
	bool NegateLeftExpr;
	int LeftBlockNum;  // SMP_BLOCKNUM_UNINIT if LeftExpr is not just a block number
	int RightBlockNum; // SMP_BLOCKNUM_UNINIT if RightExpr is not just a block number
	int NonFallThroughBlockNum;  // blocknum to execute if STARSCondExpr is true
	int FallThroughBlockNum; // blocknum to execute if STARSCondExpr is false
	STARSCFGBlock *MyCFGBlock; // parent
};

// Class to hold a skeleton of the basic block when we are collapsing blocks and performing
//  similar transformations.
class STARSCFGBlock {
public:
	// Constructors
	STARSCFGBlock(SMPFunction *Func, int BlockNumber) : BlockNum(BlockNumber), MyFunc(Func), Coalesced(false), 
		CoalescedBlockNum(SMP_BLOCKNUM_UNINIT), ShortCircuitExpr(nullptr) {
		ShortCircuitExpr = new STARSCondExpr(); // start with SMP_NULL_OPERATOR and a left block number
		ShortCircuitExpr->SetLeftExprBlockNum(BlockNumber, false);
		ShortCircuitExpr->SetCFGBlock(this);
	};

	// Get methods
	bool IsCoalesced(void) const { return Coalesced; };
	STARSCondExpr *GetExpr(void) const { return ShortCircuitExpr; };
	SMPFunction *GetFunc(void) const { return MyFunc; };
	inline int GetOriginalBlockNum(void) const { return BlockNum; };

	// Set methods
	void SetCoalesced(void) { Coalesced = true; };
	void SetExpr(STARSCondExpr *NewExpr) { ShortCircuitExpr = NewExpr; };

private:
	int BlockNum;  // Number for block ordering algorithms
	SMPFunction *MyFunc; // function containing this block
	bool Coalesced; // block has been folded into another block
	int CoalescedBlockNum; // block number this block has been folded into (which can then be folded into other blocks)
	STARSCondExpr *ShortCircuitExpr;

};

#endif
